import numpy as np


def unique_with_tolerance(arr: np.ndarray, rtol: float = 1e-5, atol: float = 1e-8) -> np.ndarray:
    """
    Find unique elements in an array with a specified tolerance.
    """
    assert arr.ndim == 2
    unique_values = {0: arr[0]}
    for idx, value in enumerate(arr[1:], start=1):
        for unique_value in unique_values.values():
            if all([np.isclose(v1, v2, rtol=rtol, atol=atol) for v1, v2 in zip(value, unique_value)]):
                break
        else:
            unique_values[idx] = value

    return np.array(list(unique_values.values()))



from aerocaps.geom.point import Point3D, Origin3D
from aerocaps.geom.vector import Vector3D
from aerocaps.units.length import Length

import pyvista as pv


__all__ = [
    "Plane",
    "PlaneX",
    "PlaneY",
    "PlaneZ"
]


class Plane:
    def __init__(self, p0: Point3D, p1: Point3D, p2: Point3D):
        """
        Creates a plane from three points

        Parameters
        ----------
        p0: Point3D
            First point (origin of the plane)
        p1: Point3D
            Second point
        p2: Point3D
            Third point
        """
        self.p0 = p0
        self.p1 = p1
        self.p2 = p2

    @property
    def origin(self) -> Point3D:
        return self.p0

    def compute_normal(self) -> Vector3D:
        """
        Computes the unit vector normal to the plane

        Returns
        -------
        Vector3D
            Unit normal vector
        """
        v01 = Vector3D(p0=self.p0, p1=self.p1)
        v02 = Vector3D(p0=self.p0, p1=self.p2)
        v_out = v01.cross(v02)
        return v_out.get_normalized_vector()

    @classmethod
    def plane_parallel_X(cls, distance_from_origin: Length):
        return cls(
            p0=Point3D(x=distance_from_origin, y=Length(m=0.0), z=Length(m=0.0)),
            p1=Point3D(x=distance_from_origin, y=Length(m=1.0), z=Length(m=0.0)),
            p2=Point3D(x=distance_from_origin, y=Length(m=0.0), z=Length(m=1.0))
        )

    @classmethod
    def plane_parallel_Y(cls, distance_from_origin: Length):
        return cls(
            p0=Point3D(x=Length(m=0.0), y=distance_from_origin, z=Length(m=0.0)),
            p1=Point3D(x=Length(m=0.0), y=distance_from_origin, z=Length(m=1.0)),
            p2=Point3D(x=Length(m=1.0), y=distance_from_origin, z=Length(m=0.0))
        )

    @classmethod
    def plane_parallel_Z(cls, distance_from_origin: Length):
        return cls(
            p0=Point3D(x=Length(m=0.0), y=Length(m=0.0), z=distance_from_origin),
            p1=Point3D(x=Length(m=1.0), y=Length(m=0.0), z=distance_from_origin),
            p2=Point3D(x=Length(m=0.0), y=Length(m=1.0), z=distance_from_origin)
        )

    def plot(self, plot: pv.Plotter, plane_kwargs: dict = None, mesh_kwargs: dict = None) -> pv.PolyData:
        """
        Plots the plane on the scene

        Parameters
        ----------
        plot:
            :obj:`pyvista.Plotter` instance
        plane_kwargs:
            Keyword arguments to pass to the :obj:`pyvista.Plane` constructor. Default: ``None``
        mesh_kwargs:
            Keyword arguments to pass to :obj:`pyvista.Plotter.add_mesh`. Default: ``None``

        Returns
        -------
        pv.PolyData
            The plane mesh object
        """
        mesh = pv.Plane(self.origin.as_array(), self.compute_normal().as_array(),
                        **plane_kwargs if plane_kwargs is not None else dict())
        plot.add_mesh(mesh, **mesh_kwargs if mesh_kwargs is not None else dict())
        return mesh


class PlaneX(Plane):
    def __init__(self):
        super().__init__(
            p0=Origin3D(),
            p1=Point3D(x=Length(m=0.0), y=Length(m=1.0), z=Length(m=0.0)),
            p2=Point3D(x=Length(m=0.0), y=Length(m=0.0), z=Length(m=1.0))
        )


class PlaneY(Plane):
    def __init__(self):
        super().__init__(
            p0=Origin3D(),
            p1=Point3D(x=Length(m=0.0), y=Length(m=0.0), z=Length(m=1.0)),
            p2=Point3D(x=Length(m=1.0), y=Length(m=0.0), z=Length(m=0.0))
        )


class PlaneZ(Plane):
    def __init__(self):
        super().__init__(
            p0=Origin3D(),
            p1=Point3D(x=Length(m=1.0), y=Length(m=0.0), z=Length(m=0.0)),
            p2=Point3D(x=Length(m=0.0), y=Length(m=1.0), z=Length(m=0.0))
        )



from abc import ABC


class Unit(ABC):
    def __init__(self, primary_unit: str):
        self.primary_unit = primary_unit

    def __add__(self, other):
        if self.__class__ == other.__class__:
            new_primary_value = getattr(self, self.primary_unit) + getattr(other, other.primary_unit)
            return self.__class__(**{self.primary_unit: new_primary_value})
        elif isinstance(other, (int, float)):
            new_primary_value = getattr(self, self.primary_unit) + other
            return self.__class__(**{self.primary_unit: new_primary_value})
        else:
            return NotImplemented

    def __sub__(self, other):
        if self.__class__ == other.__class__:
            new_primary_value = getattr(self, self.primary_unit) - getattr(other, other.primary_unit)
            return self.__class__(**{self.primary_unit: new_primary_value})
        elif isinstance(other, (int, float)):
            new_primary_value = getattr(self, self.primary_unit) - other
            return self.__class__(**{self.primary_unit: new_primary_value})
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            new_primary_value = getattr(self, self.primary_unit) * other
            return self.__class__(**{self.primary_unit: new_primary_value})
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            new_primary_value = getattr(self, self.primary_unit) / other
            return self.__class__(**{self.primary_unit: new_primary_value})
        elif self.__class__ == other.__class__:
            new_primary_value = getattr(self, self.primary_unit) / getattr(other, other.primary_unit)
            return new_primary_value
        else:
            return NotImplemented

    def __abs__(self):
        new_primary_value = abs(getattr(self, self.primary_unit))
        return self.__class__(**{self.primary_unit: new_primary_value})

    def __neg__(self):
        new_primary_value = -(getattr(self, self.primary_unit))
        return self.__class__(**{self.primary_unit: new_primary_value})



from aerocaps.iges import global_section_col_width


class StartSection:
    def __init__(self, n_start_lines: int = 1):
        self.n_start_lines = n_start_lines

    def write_start_section_string(self):
        start_section_string = ""
        for start_idx in range(self.n_start_lines):
            start_section_string += " " * global_section_col_width
            start_section_string += f"S{start_idx + 1:7d}\n"
        return start_section_string


class EndSection:
    def __init__(self, n_start_lines, n_global_lines, n_entity_lines, n_data_lines, n_end_lines: int = 1):
        self.n_start_lines = n_start_lines
        self.n_global_lines = n_global_lines
        self.n_entity_lines = n_entity_lines
        self.n_data_lines = n_data_lines
        self.n_end_lines = n_end_lines

    def write_end_section_string(self):
        end_section_string = \
            f"S{self.n_start_lines:7d}G{self.n_global_lines:7d}D{self.n_entity_lines:7d}P{self.n_data_lines:7d}"
        end_section_string += " " * (global_section_col_width - len(end_section_string))
        end_section_string += f"T{self.n_end_lines:7d}\n"
        return end_section_string



import typing

import numpy as np

from aerocaps.geom.point import Point3D, Origin3D, Point2D, Origin2D
from aerocaps.units.length import Length
from aerocaps.units.unit import Unit


__all__ = [
    "Vector2D",
    "Vector3D",
    "IHat2D",
    "JHat2D",
    "IHat3D",
    "JHat3D",
    "KHat3D"
]


class Vector2D:
    def __init__(self, p0: Point2D, p1: Point2D):
        self.p0 = p0
        self.p1 = p1

    def value(self):
        return [self.p1.x - self.p0.x, self.p1.y - self.p0.y]

    def normalized_value(self) -> typing.List[float]:
        mag = self.mag()
        return [xy / mag for xy in self.value()]

    def get_normalized_vector(self) -> "Vector2D":
        return Vector2D.from_array(np.array(self.normalized_value()))

    def scale(self, factor: float) -> "Vector2D":
        return Vector2D(self.p0, factor * (self.p1 - self.p0))

    def as_array(self):
        return np.array([xyz.m for xyz in self.value()])

    @classmethod
    def from_array(cls, arr: np.ndarray):
        return cls(p0=Origin2D(), p1=Point2D.from_array(arr))

    def is_zero_vector(self) -> bool:
        return np.all(np.isclose(self.as_array(), 0.0))

    def dot(self, other: "Vector2D") -> float or Unit:
        A = self.value()
        B = other.value()
        return A[0] * B[0] + A[1] * B[1]

    def mag(self) -> Length:
        return Length(m=np.sqrt(np.sum(self.as_array()**2)))

    def __mul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Vector2D.from_array(other * self.as_array())
        raise NotImplementedError(f"Vector scaling is currently only implemented for int and float data")

    def __rmul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Vector2D.from_array(other * self.as_array())
        raise NotImplementedError(f"Vector scaling is currently only implemented for int and float data")


class Vector3D:
    def __init__(self, p0: Point3D, p1: Point3D):
        self.p0 = p0
        self.p1 = p1

    def value(self):
        return [self.p1.x - self.p0.x, self.p1.y - self.p0.y, self.p1.z - self.p0.z]

    def normalized_value(self) -> typing.List[float]:
        mag = self.mag()
        return [xyz / mag for xyz in self.value()]

    def get_normalized_vector(self) -> "Vector3D":
        return Vector3D.from_array(np.array(self.normalized_value()))

    def scale(self, factor: float) -> "Vector3D":
        return Vector3D(self.p0, factor * (self.p1 - self.p0))

    def as_array(self):
        return np.array([xyz.m for xyz in self.value()])

    def is_zero_vector(self) -> bool:
        return np.all(np.isclose(self.as_array(), 0.0))

    @classmethod
    def from_array(cls, arr: np.ndarray):
        return cls(p0=Origin3D(), p1=Point3D.from_array(arr))

    def cross(self, other: "Vector3D") -> "Vector3D":
        return Vector3D(p0=Origin3D(),
                        p1=Point3D.from_array(
                            np.cross(self.as_array(), other.as_array())
                        ))

    def dot(self, other: "Vector3D") -> float or Unit:
        A = self.value()
        B = other.value()
        return A[0] * B[0] + A[1] * B[1] + A[2] * B[2]

    def mag(self) -> Length:
        return Length(m=np.sqrt(np.sum(self.as_array()**2)))

    def __mul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Vector3D.from_array(other * self.as_array())
        raise NotImplementedError(f"Vector scaling is currently only implemented for int and float data")

    def __rmul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            return Vector3D.from_array(other * self.as_array())
        raise NotImplementedError(f"Vector scaling is currently only implemented for int and float data")


class IHat2D(Vector2D):
    def __init__(self):
        super().__init__(p0=Origin2D(), p1=Point2D.from_array(np.array([1.0, 0.0])))


class JHat2D(Vector2D):
    def __init__(self):
        super().__init__(p0=Origin2D(), p1=Point2D.from_array(np.array([0.0, 1.0])))


class IHat3D(Vector3D):
    def __init__(self):
        super().__init__(p0=Origin3D(), p1=Point3D.from_array(np.array([1.0, 0.0, 0.0])))


class JHat3D(Vector3D):
    def __init__(self):
        super().__init__(p0=Origin3D(), p1=Point3D.from_array(np.array([0.0, 1.0, 0.0])))


class KHat3D(Vector3D):
    def __init__(self):
        super().__init__(p0=Origin3D(), p1=Point3D.from_array(np.array([0.0, 0.0, 1.0])))



import numpy as np

from aerocaps.iges.entity import IGESEntity
from aerocaps.iges.iges_param import IGESParam
from aerocaps.units.angle import Angle
from aerocaps.units.length import Length


class TransformationMatrixIGES(IGESEntity):
    def __init__(self,
                 yaw_z: Angle = Angle(deg=0.0),
                 pitch_y: Angle = Angle(deg=0.0),
                 roll_x: Angle = Angle(deg=0.0),
                 tx: Length = Length(m=0.0),
                 ty: Length = Length(m=0.0),
                 tz: Length = Length(m=0.0)
                 ):
        R_mat = self.generate_roll_pitch_yaw_matrix(yaw_z, pitch_y, roll_x)
        parameter_data = [
            IGESParam(R_mat[0, 0], "real"),
            IGESParam(R_mat[0, 1], "real"),
            IGESParam(R_mat[0, 2], "real"),
            IGESParam(tx.m, "real"),
            IGESParam(R_mat[1, 0], "real"),
            IGESParam(R_mat[1, 1], "real"),
            IGESParam(R_mat[1, 2], "real"),
            IGESParam(ty.m, "real"),
            IGESParam(R_mat[2, 0], "real"),
            IGESParam(R_mat[2, 1], "real"),
            IGESParam(R_mat[2, 2], "real"),
            IGESParam(tz.m, "real")
        ]
        super().__init__(124, parameter_data)

    @staticmethod
    def _generate_matrix_z(alpha: Angle):
        """
        Generates the yaw matrix
        """
        return np.array([
            [np.cos(alpha.rad), -np.sin(alpha.rad), 0.0],
            [np.sin(alpha.rad), np.cos(alpha.rad), 0.0],
            [0.0, 0.0, 1.0]
        ])

    @staticmethod
    def _generate_matrix_y(beta: Angle):
        """
        Generates the pitch matrix
        """
        return np.array([
            [np.cos(beta.rad), 0.0, np.sin(beta.rad)],
            [0.0, 1.0, 0.0],
            [-np.sin(beta.rad), 0.0, np.cos(beta.rad)]
        ])

    @staticmethod
    def _generate_matrix_x(gamma: Angle):
        """
        Generates the roll matrix
        """
        return np.array([
            [1.0, 0.0, 0.0],
            [0.0, np.cos(gamma.rad), -np.sin(gamma.rad)],
            [0.0, np.sin(gamma.rad), np.cos(gamma.rad)]
        ])

    def generate_roll_pitch_yaw_matrix(self, alpha: Angle, beta: Angle, gamma: Angle):
        yaw_matrix = self._generate_matrix_z(alpha)
        pitch_matrix = self._generate_matrix_y(beta)
        roll_matrix = self._generate_matrix_x(gamma)
        return (yaw_matrix @ pitch_matrix) @ roll_matrix



from datetime import datetime


class IGESParam:
    def __init__(self, value, dtype: str):
        self.value = value
        self.dtype = dtype
        allowed_dtypes = ["string", "int", "pointer", "real", "datetime", "none"]
        if self.dtype == "datetime" and not isinstance(self.value, datetime):
            raise TypeError(f"datetime was selected as the dtype for IGESParam with value {self.value}, but "
                            f"the type was {type(self.value)}. 'value' must be of type datetime.datetime.")
        if self.dtype not in allowed_dtypes:
            raise ValueError(f"GlobalParam dtype must be one of {allowed_dtypes}. Chosen value was {self.dtype}.")

    def write_value_to_python_str(self):
        if self.dtype == "int":
            return str(int(self.value))
        elif self.dtype == "pointer":
            return str(int(self.value))
        elif self.dtype == "real":
            return f"{self.value:.12E}"
        elif self.dtype == "string":
            return f"{len(self.value)}H{self.value}"  # Hollerith format string
        elif self.dtype == "datetime":
            return f"15H{self.value.strftime('%Y%m%d.%H%M%S')}"
        elif self.dtype == "none":
            return ""



import numpy as np
from scipy.optimize import minimize

import aerocaps as ac
from aerocaps.examples.bezier_surface import bezier_surface_2x3
from aerocaps.iges.iges_generator import IGESGenerator


def test_one_edge():
    surf_1 = bezier_surface_2x3()
    points = np.array([
        [
            [0.0, 0.0, 0.0],
            [0.3, 0.0, 0.1],
            [0.6, 0.0, -0.1],
            [1.0, 0.0, 0.0]
        ],
        [
            [0.0, -0.2, 0.1],
            [0.3, -0.3, 0.3],
            [0.6, -0.2, -0.1],
            [1.0, -0.075, -0.075]
        ],
        [
            [0.0, -0.4, 0.2],
            [0.3, -0.5, 0.2],
            [0.6, -0.3, 0.1],
            [1.0, -0.15625, -0.125]
        ]
    ])
    surf_2 = ac.BezierSurface(points)

    points = np.array([
        [
            [1.0, 0.0, 0.0],
            [1.0, -0.075, -0.075],
            [1.0, -0.15625, -0.125]
        ],
        [
            [1.2, 0.0, -0.1],
            [1.2, -0.075, -0.1],
            [1.25, -0.15625, -0.2]
        ],
        [
            [1.4, 0.0, -0.15],
            [1.4, -0.075, -0.15],
            [1.35, -0.15625, -0.25]
        ]
    ])
    surf_3 = ac.BezierSurface(points)

    n_points = 10
    d_g1_1 = surf_1.get_first_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
    d_g2_1 = surf_1.get_second_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
    d_g1_3 = surf_3.get_first_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
    d_g2_3 = surf_3.get_second_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)

    def optimize():
        f1 = 0.3
        f2 = 1.0
        # x0_g1 = surf_2.get_control_point_array()[1, :-1, :].flatten()
        # x0_g1_b = surf_2.get_control_point_array()[2, 2, :].flatten()
        # x0_g1 = np.append(x0_g1, x0_g1_b)
        # x0_g2 = surf_2.get_control_point_array()[2, :-2, :].flatten()

        x0 = surf_2.get_control_point_array()[1:, :-1, :].flatten()

        # def obj_fun_g1(x):
        #     x_reshaped = x.reshape((4, 3))
        #     surf_2.points[1][0] = ac.Point3D.from_array(x_reshaped[0])
        #     surf_2.points[1][1] = ac.Point3D.from_array(x_reshaped[1])
        #     surf_2.points[1][2] = ac.Point3D.from_array(x_reshaped[2])
        #     surf_2.points[2][2] = ac.Point3D.from_array(x_reshaped[3])
        #     d_g1_2 = surf_2.get_first_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
        #     d_g1_2b = surf_2.get_first_derivs_along_edge(ac.SurfaceEdge.v1, n_points=n_points)
        #     return np.sum((d_g1_1 + 1/f1 * d_g1_2)**2) + np.sum((d_g1_3 - 1/f2 * d_g1_2b)**2)
        #
        # def obj_fun_g2(x):
        #     x_reshaped = x.reshape((2, 3))
        #     surf_2.points[2][0] = ac.Point3D.from_array(x_reshaped[0])
        #     surf_2.points[2][1] = ac.Point3D.from_array(x_reshaped[1])
        #     d_g2_2 = surf_2.get_second_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
        #     d_g2_2b = surf_2.get_second_derivs_along_edge(ac.SurfaceEdge.v1, n_points=n_points)
        #     return np.sum((d_g2_1 + 1/f1**2 * d_g2_2)**2) + np.sum((d_g2_3 - 1/f2**2 * d_g2_2b)**2)

        # res_g1 = minimize(obj_fun_g1, x0_g1, method="SLSQP")
        # res_g2 = minimize(obj_fun_g2, x0_g2, method="SLSQP")
        # print(f"{res_g1 = }")
        # print(f"{res_g2 = }")

        def obj_fun(x):
            x_reshaped = x.reshape((2, 3, 3))
            for i in range(x_reshaped.shape[0]):
                for j in range(x_reshaped.shape[1]):
                    surf_2.points[i + 1][j] = ac.Point3D.from_array(x_reshaped[i, j])
            d_g1_2 = surf_2.get_first_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
            d_g1_2b = surf_2.get_first_derivs_along_edge(ac.SurfaceEdge.v1, n_points=n_points)
            d_g2_2 = surf_2.get_second_derivs_along_edge(ac.SurfaceEdge.u0, n_points=n_points)
            d_g2_2b = surf_2.get_second_derivs_along_edge(ac.SurfaceEdge.v1, n_points=n_points)
            J1 = np.sum((d_g1_1 + 1/f1 * d_g1_2)**2)
            J2 = np.sum((d_g1_3 - 1/f2 * d_g1_2b)**2)
            J3 = np.sum((d_g2_1 + 1/f1**2 * d_g2_2)**2)
            J4 = np.sum((d_g2_3 - 1/f2**2 * d_g2_2b)**2)
            return J1 + J2 + J3 + J4

        res = minimize(obj_fun, x0)
        print(f"{res = }")

    optimize()

    # iges_generator = IGESGenerator([surf_1.to_iges(), surf_2.to_iges(), surf_3.to_iges()], units="meters")
    # iges_generator.generate("surface_optimizer.igs")

    # import pyvista as pv
    # plot = pv.Plotter()
    # surf_1.plot_surface(plot, color="blue")
    # surf_2.plot_surface(plot, color="red")
    # surf_2.plot_control_points(plot, color="black", point_size=20, render_points_as_spheres=True)
    # surf_2.plot_control_point_mesh_lines(plot, color="orange")
    # surf_3.plot_surface(plot, color="green")
    # plot.show()


if __name__ == "__main__":
    test_one_edge()



import os

import matplotlib.pyplot as plt

from aerocaps.examples.rational_bezier_curve import quarter_circle_rational_bezier
from aerocaps.utils.plotting import FONT_SERIF_14, FONT_SERIF_16, format_axis_basic


def elevate_quarter_circle(image_dir: str = None):
    r"""
    Plots quarter circle exactly represented both by a quadratic rational Bézier curve and a cubic
    rational Bézier curve

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    curve = quarter_circle_rational_bezier()
    elevated = curve.elevate_degree()

    # Generate the plots
    for c in [curve, elevated]:
        n = c.degree
        fig, ax = plt.subplots()
        format_axis_basic(ax, FONT_SERIF_14)
        ax.set_aspect("equal")
        c.plot(ax, projection="XY", color="steelblue")
        c.plot_control_points(ax, projection="XY", marker="o", mfc="black", mec="black", ls=":", color="black")
        ax.text(0.4, 0.4, f"{n = }", fontdict=FONT_SERIF_16)
        ax.set_xlabel(r"$x$", font=FONT_SERIF_16)
        ax.set_ylabel(r"$y$", font=FONT_SERIF_16)
        if image_dir:
            fig.savefig(os.path.join(image_dir, f"quarter_circle_n={n}.png"), bbox_inches="tight")
        else:
            plt.show()


def main(*args, **kwargs):
    elevate_quarter_circle(*args, **kwargs)


if __name__ == "__main__":
    main(image_dir=r"C:\Users\mlaue\Documents\aerocaps\docs\source\images")



import typing

import numpy as np

from aerocaps.geom import Surface


class STLGenerator:
    """
    Reference: https://www.loc.gov/preservation/digital/formats/fdd/fdd000506.shtml
    """
    def __init__(self, geoms: typing.List[Surface], Nu: int = 50, Nv: int = 50):
        self.geoms = geoms
        self.Nu = Nu
        self.Nv = Nv

    def generate(self, file_name: str):
        with open(file_name, "w") as stl_file:
            stl_file.write("solid aerocaps\n")

            for geom in self.geoms:
                point_array = geom.evaluate_grid(self.Nu, self.Nv)
                for i in range(point_array.shape[0] - 1):
                    for j in range(point_array.shape[1] - 1):
                        vertex_1 = point_array[i, j, :]
                        vertex_2 = point_array[i + 1, j, :]
                        vertex_3 = point_array[i, j + 1, :]
                        vertex_4 = point_array[i + 1, j + 1, :]
                        normal_1 = np.cross(vertex_2 - vertex_1, vertex_3 - vertex_1)
                        normal_2 = np.cross(vertex_3 - vertex_4, vertex_2 - vertex_4)
                        stl_file.write(f"facet normal {normal_1[0]} {normal_1[1]} {normal_1[2]}\n")
                        stl_file.write(f"    outer loop\n")
                        stl_file.write(f"        vertex {vertex_1[0]} {vertex_1[1]} {vertex_1[2]}\n")
                        stl_file.write(f"        vertex {vertex_2[0]} {vertex_2[1]} {vertex_2[2]}\n")
                        stl_file.write(f"        vertex {vertex_3[0]} {vertex_3[1]} {vertex_3[2]}\n")
                        stl_file.write(f"    endloop\n")
                        stl_file.write(f"endfacet\n")
                        stl_file.write(f"facet normal {normal_2[0]} {normal_2[1]} {normal_2[2]}\n")
                        stl_file.write(f"    outer loop\n")
                        stl_file.write(f"        vertex {vertex_4[0]} {vertex_4[1]} {vertex_4[2]}\n")
                        stl_file.write(f"        vertex {vertex_2[0]} {vertex_2[1]} {vertex_2[2]}\n")
                        stl_file.write(f"        vertex {vertex_3[0]} {vertex_3[1]} {vertex_3[2]}\n")
                        stl_file.write(f"    endloop\n")
                        stl_file.write(f"endfacet\n")

            stl_file.write("endsolid aerocaps\n")



import os

import numpy as np
import copy

from aerocaps.geom.point import Point3D
from aerocaps.geom.surfaces import NURBSSurface, BezierSurface, RationalBezierSurface, SurfaceEdge, BSplineSurface
from aerocaps.geom.curves import BezierCurve3D,Line3D
from aerocaps.geom import NegativeWeightError
from aerocaps.units.angle import Angle
from rust_nurbs import *


def test_nurbs_revolve():
    axis = Line3D(p0=Point3D.from_array(np.array([0.0, 0.0, 0.0])),
                  p1=Point3D.from_array(np.array([0.0, 0.0, 1.0])))
    cubic_bezier_cps = np.array([
        [0.0, -1.0, 0.0],
        [0.0, -1.2, 0.5],
        [0.0, -1.3, 1.0],
        [0.0, -0.8, 1.5]
    ])
    bezier = BezierCurve3D([Point3D.from_array(p) for p in cubic_bezier_cps])
    nurbs_surface = NURBSSurface.from_bezier_revolve(bezier, axis, Angle(deg=15.0), Angle(deg=130.0))

    point_array = nurbs_surface.evaluate_grid(30, 30)
    for point in point_array[0, :, :]:
        radius = np.sqrt(point[0] ** 2 + point[1] ** 2)
        assert np.isclose(radius, 1.0, 1e-10)
    for point in point_array[-1, :, :]:
        radius = np.sqrt(point[0] ** 2 + point[1] ** 2)
        assert np.isclose(radius, 0.8, 1e-10)


def test_bezier_surface_1():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``BezierSurface``s.
    """
    # FOR TESTING 4x4 and 4x4 first
    n = 4
    m = 4
    num_samples = 50
    rng = np.random.default_rng(seed=42)

    cp_sets_1 = rng.random((num_samples, n+1, m+1, 3))
    cp_sets_2 = rng.random((num_samples, n+1, m+1, 3))

    #Loop through different sides of the 4x4
    
    for i in range(4):
        for j in range(4):
            side_self=SurfaceEdge(i)
            side_other=SurfaceEdge(j)



            # Loop through each pair of control point meshes
            for cp_set1, cp_set2 in zip(cp_sets_1, cp_sets_2):
                bez_surf_1 = BezierSurface(cp_set1)
                bez_surf_2 = BezierSurface(cp_set2)

                # Enforce G0, G1, and G2 continuity
                bez_surf_1.enforce_g0g1g2(bez_surf_2, 1.0, side_self, side_other)

                # Verify G0, G1, and G2 continuity
                bez_surf_1.verify_g0(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g1(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g2(bez_surf_2, side_self, side_other)



def test_bezier_surface_2():
    """
    Tests the continuity enforcement method across many random pairs of randomly sized Bezier Surfaces for the parallel degree verification
    """
    for n in range(50):
        # GENERATE THE control point arrays by randomly making a 3 element array
        random_array = np.random.randint(low=4, high=15, size=3)

        #Pick the control points randomly from the 3 element array. 
        n1 = random_array[np.random.randint(0, len(random_array) )]
        m1 = random_array[np.random.randint(0, len(random_array) )]
        
        n1m1_array=np.array([n1,m1])
        random_value=np.random.randint(0,2)
        if random_value==0:
            n2= n1m1_array[np.random.randint(0,2)]
            m2= random_array[np.random.randint(0, len(random_array) )]
        else:
            m2= n1m1_array[np.random.randint(0,2)]
            n2= random_array[np.random.randint(0, len(random_array) )]



        
        rng = np.random.default_rng(seed=42)

        cp_1 = rng.random(( n1+1, m1+1, 3))
        cp_2 = rng.random(( n2+1, m2+1, 3))

        #Loop through different compatible sides

        if (np.shape(cp_1)[0]==np.shape(cp_2)[0]):
            i_vals=np.array([0,1])
            j_vals=np.array([0,1])

        elif (np.shape(cp_1)[0]==np.shape(cp_2)[1]):
            i_vals=np.array([0,1])
            j_vals=np.array([2,3])

        elif (np.shape(cp_1)[1]==np.shape(cp_2)[0]):
            i_vals=np.array([2,3])
            j_vals=np.array([0,1])
        
        elif (np.shape(cp_1)[1]==np.shape(cp_2)[1]):
            i_vals=np.array([2,3])
            j_vals=np.array([2,3])
        
        else:
            raise ValueError("Could not find matching degrees between the surfaces")
        
        for i in i_vals:
            for j in j_vals:
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                bez_surf_1 = BezierSurface(cp_1)
                bez_surf_2 = BezierSurface(cp_2)

                # Enforce G0, G1, and G2 continuity
                bez_surf_1.enforce_g0g1g2(bez_surf_2, 1.0, side_self, side_other)

                # Verify G0, G1, and G2 continuity
                bez_surf_1.verify_g0(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g1(bez_surf_2, side_self, side_other)
                bez_surf_1.verify_g2(bez_surf_2, side_self, side_other)


def test_bezier_surface_3():
    """
    Tests the continuity enforcement method across many random pairs of randomly sized Bezier Surfaces for verifying whether the tests raise assertion errors when surfaces are incompatible.
    """
    for n in range(50):
        n1 = np.random.randint(low=4, high=10)
        m1 = np.random.randint(low=4, high=10)
        n2 = np.random.randint(low=4, high=10)
        m2 = np.random.randint(low=4, high=10)
        
        rng = np.random.default_rng(seed=42)

        cp_1 = rng.random(( n1+1, m1+1, 3))
        cp_2 = rng.random(( n2+1, m2+1, 3))

        
        
        for i in range(4):
            for j in range(4):
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                bez_surf_1 = BezierSurface(cp_1)
                bez_surf_2 = BezierSurface(cp_2)
                
                try:
                    # Enforce G0, G1, and G2 continuity
                    bez_surf_1.enforce_g0g1g2(bez_surf_2, 1.0, side_self, side_other)

                    # Verify G0, G1, and G2 continuity
                    bez_surf_1.verify_g0(bez_surf_2, side_self, side_other)
                    bez_surf_1.verify_g1(bez_surf_2, side_self, side_other)
                    bez_surf_1.verify_g2(bez_surf_2, side_self, side_other)
                except (AssertionError, ValueError):
                    continue


def test_Rational_Bezier_Surface_1():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``RationalBezierSurface``s.
    """
    rng = np.random.default_rng(seed=42)
    negative_counter=0
    for it in range(15):
        n=rng.integers(4, 6)
        m=n
        #rng = np.random.default_rng(seed=42)

        cp_1 = np.array([[[0,0,rng.random()],[1,0,rng.random()],[2,0,rng.random()],[3,0,rng.random()]],
                         [[0,1,rng.random()],[1,1,rng.random()],[2,1,rng.random()],[3,1,rng.random()]],
                         [[0,2,rng.random()],[1,2,rng.random()],[2,2,rng.random()],[3,2,rng.random()]],
                         [[0,3,rng.random()],[1,3,rng.random()],[2,3,rng.random()],[3,3,rng.random()]]],dtype=np.float64)  

        #cp_1 =rng.random(( n+1, m+1, 3))
                 
        cp_2 =  np.array([[[0,0,rng.random()],[1,0,rng.random()],[2,0,rng.random()],[3,0,rng.random()]],
                         [[0,1,rng.random()],[1,1,rng.random()],[2,1,rng.random()],[3,1,rng.random()]],
                         [[0,2,rng.random()],[1,2,rng.random()],[2,2,rng.random()],[3,2,rng.random()]],
                         [[0,3,rng.random()],[1,3,rng.random()],[2,3,rng.random()],[3,3,rng.random()]]],dtype=np.float64)            
        cp_2[:, :, 0] += 3
        #cp_2 =rng.random(( n+1, m+1, 3))
        w_1 = rng.uniform(0.8, 1.2, size=cp_1.shape[:2])
        w_2 = rng.uniform(0.8, 1.2, size=cp_2.shape[:2])

        Rat_bez_surf_1 = RationalBezierSurface(cp_1, w_1)
        Rat_bez_surf_2 = RationalBezierSurface(cp_2, w_2)

        Rat_bez_surf_1_org=copy.deepcopy(Rat_bez_surf_1)
        Rat_bez_surf_2_org=copy.deepcopy(Rat_bez_surf_2)

        for i in range(4):
            for j in range(4):
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                Rat_bez_surf_1=copy.deepcopy(Rat_bez_surf_1_org)
                Rat_bez_surf_2=copy.deepcopy(Rat_bez_surf_2_org)

                
                
                try:
                    Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                    Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)

                    # g1_self=Rat_bez_surf_1.get_first_derivs_along_edge(side_self)
                    # g2_self=Rat_bez_surf_2.get_first_derivs_along_edge(side_other)

                    # g1_self_v2=Rat_bez_surf_1.get_first_derivs_along_edge_v2(side_self)
                    # g2_self_v2=Rat_bez_surf_2.get_first_derivs_along_edge_v2(side_other)
                    
                    #print(f"{g1_self=},{g2_self=}")
                    #print(f"{g1_self_v2=},{g2_self_v2=}")
                    Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                    Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)
                except NegativeWeightError:
                    negative_counter+=1

                    # plot= pv.Plotter()
                    # Rat_bez_surf_1_org.plot_surface(plot)
                    # Rat_bez_surf_1_org.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_1_org.plot_control_points(plot)
                    # Rat_bez_surf_2_org.plot_surface(plot)
                    # Rat_bez_surf_2_org.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_2_org.plot_control_points(plot)
                    # plot.set_background('black')
                    # plot.show()

                    # plot= pv.Plotter()
                    # Rat_bez_surf_1.plot_surface(plot)
                    # Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_1.plot_control_points(plot)
                    # Rat_bez_surf_2.plot_surface(plot)
                    # Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_2.plot_control_points(plot)
                    # plot.set_background('black')
                    # plot.show()

                    #print(f'{it=},{negative_counter=}')

                    

                #except NegativeWeightError:
    print(f"{negative_counter=}")
                #negative_counter+=1
                #continue
                # Enforce G0, G1, and G2 continuity

                # Verify G0, G1, and G2 continuity
                
                
                
    #print(f"{negative_counter=}")


def test_Rational_Bezier_Surface_2():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``RationalBezierSurface``s.
    """
    #rng = np.random.default_rng(seed=42)
    rng = np.random.default_rng(seed=60)
    Assertion_error_counter=0
    Negative_error_counter=0
    num_enforced=0
    flag=False
    for n in range(40):

        if flag==True:
            break
        

        random_array = rng.integers(low=4, high=6, size=3)

        #Pick the control points randomly from the 3 element array. 
        n1 = random_array[rng.integers(0, len(random_array) )]
        m1 = random_array[rng.integers(0, len(random_array) )]
        
        n1m1_array=np.array([n1,m1])
        random_value=rng.integers(0,2)
        if random_value==0:
            n2= n1m1_array[rng.integers(0,2)]
            m2= random_array[rng.integers(0, len(random_array) )]
        else:
            m2= n1m1_array[rng.integers(0,2)]
            n2= random_array[rng.integers(0, len(random_array) )]



        
        

        cp_1 = rng.random(( n1+1, m1+1, 3))
        #cp_1 = rng.random(( 4, 5, 3))

        # cp_1 = np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
        #                  [[0,1,1],[1,1,0],[2,1,1],[3,1,1]],
        #                  [[0,2,0],[1,2,1],[2,2,0],[3,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)  

        # cp_1 = np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1],[4,0,1]],
        #                  [[0,1,1],[1,1,0],[2,1,1],[3,1,1],[4,1,1]],
        #                  [[0,2,0],[1,2,1],[2,2,0],[3,2,1],[4,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1],[4,3,1]]],dtype=np.float64) 
        

        # cp_1 = np.array([[[0,0,rng.random()],[1,0,rng.random()],[2,0,rng.random()],[3,0,rng.random()],[4,0,rng.random()]],
        #                  [[0,1,rng.random()],[1,1,rng.random()],[2,1,rng.random()],[3,1,rng.random()],[4,1,rng.random()]],
        #                  [[0,2,rng.random()],[1,2,rng.random()],[2,2,rng.random()],[3,2,rng.random()],[4,2,rng.random()]],
        #                  [[0,3,rng.random()],[1,3,rng.random()],[2,3,rng.random()],[3,3,rng.random()],[4,3,rng.random()]]],dtype=np.float64) 

        cp_2 = rng.random(( n2+1, m2+1, 3))
        
        #cp_2 = rng.random(( 4, 5, 3))

        

        # cp_2 =  np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
        #                  [[0,1,2],[1,1,1],[2,1,1],[3,1,1]],
        #                  [[0,2,0],[1,2,0],[2,2,1],[3,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)   

        # cp_2 =  np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1],[4,0,1]],
        #                  [[0,1,2],[1,1,1],[2,1,1],[3,1,1],[4,1,1]],
        #                  [[0,2,0],[1,2,0],[2,2,1],[3,2,1],[4,2,1]],
        #                  [[0,3,0],[1,3,1],[2,3,1],[3,3,1],[4,3,1]]],dtype=np.float64)
        # cp_2[:, :, 0] += 4      

        # cp_2 =  np.array([[[0,0,rng.random()],[1,0,rng.random()],[2,0,rng.random()],[3,0,rng.random()],[4,0,rng.random()]],
        #                  [[0,1,rng.random()],[1,1,rng.random()],[2,1,rng.random()],[3,1,rng.random()],[4,1,rng.random()]],
        #                  [[0,2,rng.random()],[1,2,rng.random()],[2,2,rng.random()],[3,2,rng.random()],[4,2,rng.random()]],
        #                  [[0,3,rng.random()],[1,3,rng.random()],[2,3,rng.random()],[3,3,rng.random()],[4,3,rng.random()]]],dtype=np.float64)
        # cp_2[:, :, 0] += 4 



        # w_1 = rng.uniform(0.4, 0.5, (n1+1, m1+1))
        # w_2 = rng.uniform(0.9, 1.2, (n2+1, m2+1))

        w_1 = rng.uniform(0.8, 1.2, (np.shape(cp_1)[0], np.shape(cp_1)[1]))
        w_2 = rng.uniform(0.8, 1.3, (np.shape(cp_2)[0], np.shape(cp_2)[1]))

        # w_1[0][0]=1
        # w_1[0][-1]=1
        # w_1[-1][0]=1
        # w_1[-1][-1]=1

        # # w_1[0][:]=1
        # # w_1[:][-1]=1
        # # w_1[-1][:]=1
        # # w_1[:][-1]=1

        # w_2[0][0]=1
        # w_2[0][-1]=1
        # w_2[-1][0]=1
        # w_2[-1][-1]=1

        # w_2[0][:]=1
        # w_2[:][-1]=1
        # w_2[-1][:]=1
        # w_2[:][-1]=1

        
        # print(f'{w_1=}')
        # print(f'{w_2=}')


        
        #print(f'{cp_1=},{np.shape(cp_1)=}')
        #print(f'{cp_2=},{np.shape(cp_2)=}')
        
        #Loop through different compatible sides

        if (np.shape(cp_1)[0]==np.shape(cp_2)[0]):
            i_vals=np.array([0,1])
            j_vals=np.array([0,1])

        elif (np.shape(cp_1)[0]==np.shape(cp_2)[1]):
            i_vals=np.array([0,1])
            j_vals=np.array([2,3])

        elif (np.shape(cp_1)[1]==np.shape(cp_2)[0]):
            i_vals=np.array([2,3])
            j_vals=np.array([0,1])
        
        elif (np.shape(cp_1)[1]==np.shape(cp_2)[1]):
            i_vals=np.array([2,3])
            j_vals=np.array([2,3])
        
        
        else:
            raise ValueError("Could not find matching degrees between the surfaces")
        
        # i_vals=np.array([0,1])
        # j_vals=np.array([0,1])
        
        Rat_bez_surf_1 = RationalBezierSurface(cp_1,w_1)
        Rat_bez_surf_2 = RationalBezierSurface(cp_2,w_2)

        # plot= pv.Plotter()
        # Rat_bez_surf_1.plot_surface(plot)
        # Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
        # Rat_bez_surf_1.plot_control_points(plot)
        # Rat_bez_surf_2.plot_surface(plot)
        # Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
        # Rat_bez_surf_2.plot_control_points(plot)
        # plot.set_background('black')
        # plot.show()

        # plot= pv.Plotter()
        # Rat_bez_surf_1.plot_surface(plot)
        # Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
        # Rat_bez_surf_1.plot_control_points(plot)
        # Rat_bez_surf_2.plot_surface(plot)
        # Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
        # Rat_bez_surf_2.plot_control_points(plot)
        # plot.set_background('black')
        # plot.show()
        
        Rat_bez_surf_1_org=copy.deepcopy(Rat_bez_surf_1)
        Rat_bez_surf_2_org=copy.deepcopy(Rat_bez_surf_2)
        

        #COUNT NUMBER OF ENFORCEMENTS TRIED
        

        for i in i_vals:
            for j in j_vals:
                num_enforced=num_enforced+1
                
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                #RESET TO ORIGINAL FOR EVERY ITERATION OF LOOP

                Rat_bez_surf_1=copy.deepcopy(Rat_bez_surf_1_org)
                Rat_bez_surf_2=copy.deepcopy(Rat_bez_surf_2_org)

                


                # Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                    
                # # Verify G0, G1, and G2 continuity
                # Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)
                # Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                # Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)

                # plot= pv.Plotter()
                # Rat_bez_surf_1.plot_surface(plot)
                # Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
                # Rat_bez_surf_1.plot_control_points(plot)
                # Rat_bez_surf_2.plot_surface(plot)
                # Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
                # Rat_bez_surf_2.plot_control_points(plot)
                # plot.set_background('black')
                # plot.show()

                # Loop through each pair of control point meshes
                
                # Enforce G0, G1, and G2 continuity
                try:
                    Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                    
                    # Verify G0, G1, and G2 continuity
                    Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)
                    Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                    Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)
                except AssertionError:
                    Assertion_error_counter+=1

                    # iges_entities = [Rat_bez_surf_1.to_iges(),Rat_bez_surf_2.to_iges()]
                    # cp_net_points, cp_net_lines = Rat_bez_surf_1.generate_control_point_net()
                    # iges_entities.extend([cp_net_point.to_iges() for cp_net_point in cp_net_points])
                    # iges_entities.extend([cp_net_line.to_iges() for cp_net_line in cp_net_lines])
                    # cp_net_points_2, cp_net_lines_2 = Rat_bez_surf_2.generate_control_point_net()
                    # iges_entities.extend([cp_net_point.to_iges() for cp_net_point in cp_net_points_2])
                    # iges_entities.extend([cp_net_line.to_iges() for cp_net_line in cp_net_lines_2])

                    # #iges_file = os.path.join(TEST_DIR, "Rat_Bez_test.igs")
                    # iges_file = os.path.join(r"C:\aerocaps-main\aerocaps\aerocaps\tests", "Rat_Bez_test_5.igs")
                    # print(f"{iges_file=}")
                    # iges_generator = IGESGenerator(iges_entities, "meters")
                    # iges_generator.generate(iges_file)
                    # print("Generator passed")

                    flag=True

                    
                    
                except NegativeWeightError:
                    Negative_error_counter+=1
                    #print(f'{i=},{j=}')
                    #print(f"{side_self=},{side_other=}")
                    #print(f'{cp_1=},{cp_2=}')

                    # if (i==0 and j==1):
                    #     fail_case_1=cp_1
                    #     fail_case_2=cp_2
                    #     weight_case1=w_1
                    #     weight_case2=w_2
                    #     flag=True
                    #     break
                    # plot= pv.Plotter()
                    # Rat_bez_surf_1_org.plot_surface(plot)
                    # Rat_bez_surf_1_org.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_1_org.plot_control_points(plot)
                    # Rat_bez_surf_2_org.plot_surface(plot)
                    # Rat_bez_surf_2_org.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_2_org.plot_control_points(plot)
                    # plot.set_background('black')
                    # plot.show()

                    # plot= pv.Plotter()
                    # Rat_bez_surf_1.plot_surface(plot)
                    # Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_1.plot_control_points(plot)
                    # Rat_bez_surf_2.plot_surface(plot)
                    # Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
                    # Rat_bez_surf_2.plot_control_points(plot)
                    # plot.set_background('black')
                    # plot.show()
        
    print(f'{n=},{num_enforced=}')    
    print(f'{n=},{Assertion_error_counter=}')
    print(f'{n=},{Negative_error_counter=}')
    #print(f'{fail_case_1=},{fail_case_2=}')

    # return fail_case_1,fail_case_2,weight_case1,weight_case2

    

test_Rational_Bezier_Surface_2()
# fc1,fc2,w1,w2=test_Rational_Bezier_Surface_2()

# print(f'{fc1=},{fc2=},{w1=},{w2=}')


# Rat_bez_surf_1 = RationalBezierSurface(fc1,w1)
# Rat_bez_surf_2 = RationalBezierSurface(fc2,w2)
# side_self=SurfaceEdge(0)
# side_other=SurfaceEdge(1)

# plot= pv.Plotter()
# Rat_bez_surf_1.plot_surface(plot)
# Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
# Rat_bez_surf_1.plot_control_points(plot)
# Rat_bez_surf_2.plot_surface(plot)
# Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
# Rat_bez_surf_2.plot_control_points(plot)
# plot.set_background('black')
# plot.show()

# Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                    
# # Verify G0, G1, and G2 continuity
# # Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)
# # Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
# # Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)

# plot= pv.Plotter()
# Rat_bez_surf_1.plot_surface(plot)
# Rat_bez_surf_1.plot_control_point_mesh_lines(plot)
# Rat_bez_surf_1.plot_control_points(plot)
# Rat_bez_surf_2.plot_surface(plot)
# Rat_bez_surf_2.plot_control_point_mesh_lines(plot)
# Rat_bez_surf_2.plot_control_points(plot)
# plot.set_background('black')
# plot.show()



def test_Rational_Bezier_Surface_3():
    """
    Tests the continuity enforcement method across many random pairs of 4x4 ``RationalBezierSurface``s.
    """
    rng = np.random.default_rng(seed=42)
    for n in range(1):

        random_array = rng.integers(low=4, high=15, size=3)

        #Pick the control points randomly from the 3 element array. 
        n1 = random_array[rng.integers(0, len(random_array) )]
        m1 = random_array[rng.integers(0, len(random_array) )]
        
        n1m1_array=np.array([n1,m1])
        random_value=rng.integers(0,2)
        if random_value==0:
            n2= n1m1_array[rng.integers(0,2)]
            m2= random_array[rng.integers(0, len(random_array) )]
        else:
            m2= n1m1_array[rng.integers(0,2)]
            n2= random_array[rng.integers(0, len(random_array) )]



        
        

        #cp_1 = rng.random(( n1+1, m1+1, 3))

        cp_1 = np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
                         [[0,1,1],[1,1,0],[2,1,1],[3,1,1]],
                         [[0,2,0],[1,2,1],[2,2,0],[3,2,1]],
                         [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)  
        
        #cp_2 = rng.random(( n2+1, m2+1, 3))

        cp_2 =  np.array([[[0,0,1],[1,0,1],[2,0,1],[3,0,1]],
                         [[0,1,2],[1,1,1],[2,1,1],[3,1,1]],
                         [[0,2,0],[1,2,0],[2,2,1],[3,2,1]],
                         [[0,3,0],[1,3,1],[2,3,1],[3,3,1]]],dtype=np.float64)            

        w_1 = rng.uniform(0.9, 1.1, cp_1.shape[:2])
        w_2 = rng.uniform(0.9, 1.1, cp_2.shape[:2])

        #Loop through different compatible sides

        if (np.shape(cp_1)[0]==np.shape(cp_2)[0]):
            i_vals=np.array([0,1])
            j_vals=np.array([0,1])

        elif (np.shape(cp_1)[0]==np.shape(cp_2)[1]):
            i_vals=np.array([0,1])
            j_vals=np.array([2,3])

        elif (np.shape(cp_1)[1]==np.shape(cp_2)[0]):
            i_vals=np.array([2,3])
            j_vals=np.array([0,1])
        
        elif (np.shape(cp_1)[1]==np.shape(cp_2)[1]):
            i_vals=np.array([2,3])
            j_vals=np.array([2,3])
        
        else:
            raise ValueError("Could not find matching degrees between the surfaces")
        
        for i in i_vals:
            for j in j_vals:
                side_self=SurfaceEdge(i)
                side_other=SurfaceEdge(j)

                # Loop through each pair of control point meshes
                
                Rat_bez_surf_1 = RationalBezierSurface(cp_1,w_1)
                Rat_bez_surf_2 = RationalBezierSurface(cp_2,w_2)

                # Enforce G0, G1, and G2 continuity
                Rat_bez_surf_1.enforce_g0g1g2(Rat_bez_surf_2, 1.0, side_self, side_other)
                
                # Verify G0, G1, and G2 continuity
                Rat_bez_surf_1.verify_g0(Rat_bez_surf_2, side_self, side_other)
                Rat_bez_surf_1.verify_g1(Rat_bez_surf_2, side_self, side_other)
                Rat_bez_surf_1.verify_g2(Rat_bez_surf_2, side_self, side_other)


def test_NURBS_1():
    """
    Tests the continuity enforcement method across many semi random pairs of 5x5 ``NURBS_Surfaces``s.
    All the knots are uniform and equal for both the parallel and perpendicular degrees
    """
    # rng = np.random.default_rng(seed=42)
    rng = np.random.default_rng(seed=35)
    Assertion_error_counter = 0
    Negative_error_counter = 0
    num_tried = 0
    flag = False
    for n in range(1):

        if flag:
            break

        # cp_1 = np.array([
        #     [[0, 0, rng.random()], [1, 0, rng.random()], [2, 0, rng.random()], [3, 0, rng.random()], [4, 0, rng.random()]],
        #     [[0, 1, rng.random()], [1, 1, rng.random()], [2, 1, rng.random()], [3, 1, rng.random()], [4, 1, rng.random()]],
        #     [[0, 2, rng.random()], [1, 2, rng.random()], [2, 2, rng.random()], [3, 2, rng.random()], [4, 2, rng.random()]],
        #     [[0, 3, rng.random()], [1, 3, rng.random()], [2, 3, rng.random()], [3, 3, rng.random()], [4, 3, rng.random()]],
        #     [[0, 4, rng.random()], [1, 4, rng.random()], [2, 4, rng.random()], [3, 4, rng.random()], [4, 4, rng.random()]]
        # ], dtype=np.float64)
        #
        # cp_2 = np.array([
        #     [[0, 0, rng.random()], [1, 0, rng.random()], [2, 0, rng.random()], [3, 0, rng.random()], [4, 0, rng.random()]],
        #     [[0, 1, rng.random()], [1, 1, rng.random()], [2, 1, rng.random()], [3, 1, rng.random()], [4, 1, rng.random()]],
        #     [[0, 2, rng.random()], [1, 2, rng.random()], [2, 2, rng.random()], [3, 2, rng.random()], [4, 2, rng.random()]],
        #     [[0, 3, rng.random()], [1, 3, rng.random()], [2, 3, rng.random()], [3, 3, rng.random()], [4, 3, rng.random()]],
        #     [[0, 4, rng.random()], [1, 4, rng.random()], [2, 4, rng.random()], [3, 4, rng.random()], [4, 4, rng.random()]]
        # ], dtype=np.float64)
        # cp_2[:, :, 1] += 4.0

        cp_1 = rng.uniform(low=-4.0, high=4.0, size=(6, 6, 3))
        cp_2 = rng.uniform(low=-4.0, high=4.0, size=(6, 6, 3))

        w_1 = rng.uniform(0.9, 1.1, (np.shape(cp_1)[0], np.shape(cp_1)[1]))
        w_2 = rng.uniform(0.9, 1.1, (np.shape(cp_2)[0], np.shape(cp_2)[1]))

        # w_1 = np.ones((6, 6))
        # w_2 = np.ones((6, 6))

        # w_1[:, 0] = 1.0
        # w_1[:, -1] = 1.0
        # w_1[0, :] = 1.0
        # w_1[-1, :] = 1.0
        # w_2[:, 0] = 1.0
        # w_2[:, -1] = 1.0
        # w_2[0, :] = 1.0
        # w_2[-1, :] = 1.0

        p = 5
        m = p + 6 + 1
        u_knot = np.zeros(m)
        u_knot[:(p + 1)] = 0.0
        u_knot[-(p + 1):] = 1.0
        # u_knot[p + 1] = 0.25
        # u_knot[p + 2] = 0.5
        # u_knot[p + 3] = 0.75

        v_knot = np.zeros(m)
        v_knot[:(p + 1)] = 0.0
        v_knot[-(p + 1):] = 1.0
        # v_knot[p + 1] = 0.25
        # v_knot[p + 2] = 0.5
        # v_knot[p + 3] = 0.75

        print(f"{u_knot = }")
        print(f"{v_knot = }")

        NURBS_1 = NURBSSurface(cp_1, u_knot, v_knot, w_1)
        NURBS_2 = NURBSSurface(cp_2, u_knot, v_knot, w_2)

        NURBS_1.enforce_g0g1g2(NURBS_2, 1.0, SurfaceEdge.v0, SurfaceEdge.u0)
        NURBS_1.verify_g0(NURBS_2, SurfaceEdge.v0, SurfaceEdge.u0)
        NURBS_1.verify_g1(NURBS_2, SurfaceEdge.v0, SurfaceEdge.u0)
        NURBS_1.verify_g2(NURBS_2, SurfaceEdge.v0, SurfaceEdge.u0)

        # COMPARE TO FDM
        pts_edge = 10
        step = 1e-6
        term1_1st = np.array(nurbs_surf_eval_iso_v(NURBS_1.get_control_point_array(), NURBS_1.weights, u_knot, v_knot, pts_edge, 1.0))
        term2_1st = np.array(nurbs_surf_eval_iso_v(NURBS_1.get_control_point_array(), NURBS_1.weights, u_knot, v_knot, pts_edge, 1.0 - step))
        FDM_first_der_self_array = (term1_1st - term2_1st) / step
        term1 = np.array(nurbs_surf_eval_iso_v(NURBS_1.get_control_point_array(), NURBS_1.weights, u_knot, v_knot, pts_edge, 1.0))
        term2 = np.array(nurbs_surf_eval_iso_v(NURBS_1.get_control_point_array(), NURBS_1.weights, u_knot, v_knot, pts_edge, 1.0 - step))
        term3 = np.array(nurbs_surf_eval_iso_v(NURBS_1.get_control_point_array(), NURBS_1.weights, u_knot, v_knot, pts_edge, 1.0 - 2 * step))
        print(f"{term1 = }")
        print(f"{term2 = }")
        print(f"{term3 = }")
        FDM_second_der_self_array = (term1 - 2 * term2 + term3) / (step ** 2)
        print(f'{FDM_first_der_self_array=}')
        print(f'{FDM_second_der_self_array=}')

        print(f"{NURBS_1.get_first_derivs_along_edge(SurfaceEdge.v1, pts_edge, perp=True) = }")
        print(f"{NURBS_1.get_second_derivs_along_edge(SurfaceEdge.v1, pts_edge, perp=True) = }")

        NURBS_1.verify_g2(NURBS_2, SurfaceEdge.v0, SurfaceEdge.u0)


def test_bspline_surf():
    """
    Tests the continuity enforcement method across many semi random pairs of 5x5 ``NURBS_Surfaces``s.
    All the knots are uniform and equal for both the parallel and perpendicular degrees
    """
    # rng = np.random.default_rng(seed=42)
    rng = np.random.default_rng(seed=35)
    cp_1 = rng.uniform(low=-4.0, high=4.0, size=(6, 6, 3))
    cp_2 = rng.uniform(low=-4.0, high=4.0, size=(6, 6, 3))

    p = 2
    m = p + 6 + 1
    u_knot = np.zeros(m)
    u_knot[:(p + 1)] = 0.0
    u_knot[-(p + 1):] = 1.0
    # u_knot[p + 1] = 0.3
    # u_knot[p + 2] = 0.5
    # u_knot[p + 3] = 0.7
    u_knot[p + 1] = 0.5
    u_knot[p + 2] = 0.5
    u_knot[p + 3] = 0.5

    p = 2
    m = p + 6 + 1
    v_knot = np.zeros(m)
    v_knot[:(p + 1)] = 0.0
    v_knot[-(p + 1):] = 1.0
    v_knot[p + 1] = 0.3
    v_knot[p + 2] = 0.5
    v_knot[p + 3] = 0.7

    print(f"{u_knot = }")
    print(f"{v_knot = }")

    bspline_surf_1 = BSplineSurface(cp_1, u_knot, v_knot)
    bspline_surf_2 = BSplineSurface(cp_2, u_knot, v_knot)

    bspline_surf_1.enforce_g0g1g2(bspline_surf_2, 1.0, SurfaceEdge.v0, SurfaceEdge.v1)
    bspline_surf_1.verify_g0(bspline_surf_2, SurfaceEdge.v0, SurfaceEdge.v1)
    bspline_surf_1.verify_g1(bspline_surf_2, SurfaceEdge.v0, SurfaceEdge.v1)

    # TODO: understand why this next verification does not pass
    # bspline_surf_1.verify_g2(bspline_surf_2, SurfaceEdge.v0, SurfaceEdge.v1)



from aerocaps.geom.curves import Line3D
from aerocaps.geom.plane import Plane
from aerocaps.geom.point import Point3D
from aerocaps.geom.vector import Vector3D


__all__ = [
    "intersection_of_line_and_plane"
]


def intersection_of_line_and_plane(line: Line3D, plane: Plane):
    p01 = Vector3D(p0=plane.p0, p1=plane.p1)
    p02 = Vector3D(p0=plane.p0, p1=plane.p2)
    cross_vec = p01.cross(p02)
    l_a_p0 = Vector3D(p0=plane.p0, p1=line.p0)
    l_ab = Vector3D(p0=line.p0, p1=line.p1)
    numerator = cross_vec.dot(l_a_p0)
    denominator = -(l_ab.dot(cross_vec))
    t = numerator / denominator
    l_ab_val = l_ab.value()
    l_ab_t = Point3D(x=l_ab_val[0] * t, y=l_ab_val[1] * t, z=l_ab_val[2] * t)
    return line.p0 + l_ab_t



import numpy as np

import aerocaps as ac


def quarter_circle_rational_bezier() -> ac.RationalBezierCurve3D:
    points = np.array([
        [0.0, 1.0, 0.0],
        [1.0, 1.0, 0.0],
        [1.0, 0.0, 0.0]
    ])
    weights = np.array([1.0, 1 / np.sqrt(2.0), 1.0])

    curve = ac.RationalBezierCurve3D.generate_from_array(points, weights)
    return curve



import numpy as np

from aerocaps.iges.entity import IGESEntity
from aerocaps.iges.iges_param import IGESParam


class PointIGES(IGESEntity):
    def __init__(self, xyz: np.ndarray):
        """
        IGES IGESEntity #116
        """
        parameter_data = [
            IGESParam(xyz[0], "real"),
            IGESParam(xyz[1], "real"),
            IGESParam(xyz[2], "real"),
            IGESParam(0, "pointer")
        ]
        super().__init__(116, parameter_data)



import numpy as np

import aerocaps as ac


def bezier_surface_2x3() -> ac.BezierSurface:
    r"""
    Creates a Bézier surface with degrees :math:`n=2` (three rows of control points in the :math:`u`-direction)
    and :math:`m=3` (four rows of control points in the :math:`v`-direction).

    Returns
    -------
    BezierSurface
        The :math:`2 \times 3` Bézier surface
    """
    # Create three rows of control points that each influence the surface shape in the v-direction
    row_1 = np.array([
        [0.0, 0.0, 0.0],
        [0.3, 0.0, 0.1],
        [0.6, 0.0, -0.1],
        [1.0, 0.0, 0.0]
    ])
    row_2 = np.array([
        [0.0, 0.35, 0.0],
        [0.3, 0.25, 0.2],
        [0.6, 0.35, -0.2],
        [1.0, 0.30, 0.3]
    ])
    row_3 = np.array([
        [0.0, 0.7, 0.2],
        [0.3, 0.6, -0.1],
        [0.6, 0.65, 0.3],
        [1.0, 0.7, 0.2]
    ])

    # Stack the rows into a matrix
    control_points = np.stack((row_1, row_2, row_3), axis=0)

    # Generate the surface from the array
    surf = ac.BezierSurface(control_points)

    return surf



import typing

import numpy as np

import aerocaps as ac


def fill_surface_four_sided(displacement_degree: int = 3) -> (ac.RationalBezierSurface, typing.List[ac.BezierCurve3D or ac.RationalBezierCurve3D]):
    """
    Creates a four-sided fill surface from a combination of rational and non-rational Bézier curves.

    .. figure:: ../images/fill_surface.*
        :width: 600
        :align: center

        Fill surface from four curve boundaries

    Parameters
    ----------
    displacement_degree: int
        Displacement degree to use for the fill surface top and bottom boundary insertion

    Returns
    -------
    ac.RationalBezierSurface, typing.List[ac.Bezier3D or ac.RationalBezierCurve3D]
        Fill surface and the list of the four boundary curves
    """
    left_curve = ac.RationalBezierCurve3D(
        np.array([
            [0.0, 0.0, 0.0],
            [0.5, 0.5, 0.0],
            [0.0, 1.0, 0.0]
        ]),
        np.array([1.0, 0.7, 1.0])
    )
    right_curve = ac.BezierCurve3D(
        np.array([
            [2.0, 0.1, 0.0],
            [2.1, 0.9, 0.0]
        ])
    )
    top_curve = ac.BezierCurve3D(
        np.array([
            left_curve.control_points[-1].as_array(),
            [0.5, 1.2, 0.0],
            [1.5, 0.8, 0.0],
            right_curve.control_points[-1].as_array()
        ])
    )
    bottom_curve = ac.BezierCurve3D(
        np.array([
            left_curve.control_points[0].as_array(),
            [1.2, 0.2, 0.0],
            right_curve.control_points[0].as_array()
        ])
    )
    return (
        ac.RationalBezierSurface.fill_surface_from_four_boundaries(left_curve, right_curve, top_curve, bottom_curve,
                                                                   displacement_degree=displacement_degree),
        [left_curve, right_curve, top_curve, bottom_curve]
    )



import os

import numpy as np
import pyvista as pv

import aerocaps as ac
import aerocaps.examples.bezier_surface
import aerocaps.examples.fill_surface


def plot_bezier_surf_2x3(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(2 \times 3)` Bézier surface along with the control point net.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5)
    surf.plot_control_points(plot, render_points_as_spheres=True, point_size=20, color="black")
    surf.plot_control_point_mesh_lines(plot, color="#d6693a")
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_2x3.png"), scale=1)
    else:
        plot.show()


def plot_bezier_surf_2x3_mesh_only(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(2 \times 3)` Bézier surface with the isoparametric curves.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5, show_edges=True)
    plot.add_arrows(np.array([[0.1, 0.1, 0.05], [0.1, 0.1, 0.05]]), np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]),
                    mag=0.2, color="black")
    plot.add_point_labels(np.array([[0.1, 0.32, 0.05], [0.32, 0.1, 0.05]]), ["u", "v"], font_size=32)
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_2x3_mesh_only.png"), scale=1)
    else:
        plot.show()


def plot_bezier_surf_2x3_uv_labels(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(2 \times 3)` Bézier surface along with the control point net.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5)
    surf.plot_control_points(plot, render_points_as_spheres=True, point_size=20, color="black")
    surf.plot_control_point_mesh_lines(plot, color="#d6693a")
    plot.add_arrows(np.array([[0.05, 0.05, 0.05], [0.05, 0.05, 0.05]]), np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]),
                    mag=0.2, color="black")
    plot.add_point_labels(np.array([[0.1, 0.32, 0.05], [0.32, 0.1, 0.05]]), ["u", "v"], font_size=32)
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_2x3_uv_labels.png"), scale=1)
    else:
        plot.show()


def plot_bezier_surf_2x3_u_elevated(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(3 \times 3)` Bézier surface created by degree elevation from a
    :math:`2 \times 3` surface.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()
    surf = surf.elevate_degree_u()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5)
    surf.plot_control_points(plot, render_points_as_spheres=True, point_size=20, color="black")
    surf.plot_control_point_mesh_lines(plot, color="#d6693a")
    plot.add_arrows(np.array([[0.05, 0.05, 0.05], [0.05, 0.05, 0.05]]), np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]),
                    mag=0.2, color="black")
    plot.add_point_labels(np.array([[0.1, 0.32, 0.05], [0.32, 0.1, 0.05]]), ["u", "v"], font_size=32)
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_3x3_from_u_elevation.png"), scale=1)
    else:
        plot.show()


def plot_bezier_surf_2x3_v_elevated(image_dir: str = None):
    r"""
    Plots an :math:`(n \times m)=(2 \times 4)` Bézier surface created by degree elevation from a
    :math:`2 \times 3` surface.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf = aerocaps.examples.bezier_surface.bezier_surface_2x3()
    surf = surf.elevate_degree_v()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5)
    surf.plot_control_points(plot, render_points_as_spheres=True, point_size=20, color="black")
    surf.plot_control_point_mesh_lines(plot, color="#d6693a")
    plot.add_arrows(np.array([[0.05, 0.05, 0.05], [0.05, 0.05, 0.05]]), np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 0.0]]),
                    mag=0.2, color="black")
    plot.add_point_labels(np.array([[0.1, 0.32, 0.05], [0.32, 0.1, 0.05]]), ["u", "v"], font_size=32)
    plot.camera.position = [-0.5, -0.7, 0.5]
    plot.camera.focal_point = [0.6, 0.5, -0.05]

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_surf_2x4_from_v_elevation.png"), scale=1)
    else:
        plot.show()


def plot_fill_surface(image_dir: str = None):
    r"""
    Plots a fill surface along with its boundary curves.

    Parameters
    ----------
    image_dir: str or None
        Where to store the output image. If no location is specified, the image file will not be saved, and the
        plot will be displayed on-screen instead. Default: ``None``
    """
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])

    # Load the example surface
    surf, boundaries = aerocaps.examples.fill_surface.fill_surface_four_sided()

    surf.plot_surface(plot, color="#4d86b8", opacity=0.5)
    for curve in boundaries:
        curve.plot(plot, color="black")
    # surf.plot_control_points(plot, render_points_as_spheres=True, point_size=20, color="black")
    # surf.plot_control_point_mesh_lines(plot, color="#d6693a")
    plot.camera.position = [-1.2, -0.9, 1.0]
    plot.camera.focal_point = [0.6, 0.5, -0.05]
    plot.camera.zoom(1.2)

    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "fill_surface.png"), scale=1)
    else:
        plot.show()


def plot_enforce_g0g1_multiface(image_dir: str = None):
    bez_surf_u0 = ac.BezierSurface(np.array([
        [
            [0.0, 0.0, 0.0],
            [0.3, 0.1, 0.1],
            [0.6, -0.1, 0.0],
            [1.0, 0.0, 0.0]
        ],
        [
            [0.0, 0.5, 0.0],
            [0.3, 0.7, 0.1],
            [0.6, 0.4, 0.1],
            [1.0, 0.5, 0.0]
        ],
        [
            [0.0, 1.0, 0.0],
            [0.3, 1.1, 0.2],
            [0.6, 0.9, 0.1],
            [1.0, 1.0, 0.0]
        ]
    ]))
    bez_surf_u1 = ac.BezierSurface(np.array([
        [
            [0.0, -1.0, 0.0],
            [0.3, -0.7, 0.1],
            [0.6, -1.0, 0.0],
            [1.0, -1.0, 0.0]
        ],
        [
            [0.0, -1.5, 0.0],
            [0.3, -1.3, 0.2],
            [0.6, -1.4, 0.2],
            [1.0, -1.5, 0.0]
        ],
        [
            [0.0, -2.0, 0.0],
            [0.3, -1.9, 0.2],
            [0.6, -2.1, 0.0],
            [1.0, -2.0, 0.0]
        ]
    ]))
    bez_surf_v0 = ac.BezierSurface(np.array([
        [
            [0.0, 0.0, 0.0],
            [-0.1, -0.3, 0.0],
            [0.1, -0.8, 0.1],
            [0.0, -1.0, 0.0]
        ],
        [
            [-1.0, 0.0, 0.0],
            [-1.1, -0.4, 0.1],
            [-0.9, -0.8, 0.0],
            [-1.0, -1.1, 0.0]
        ],
        [
            [-2.0, 0.0, 0.0],
            [-2.0, -0.3, 0.0],
            [-2.0, -0.7, 0.0],
            [-2.0, -1.0, 0.0]
        ]
    ]))
    bez_surf_v1 = ac.BezierSurface(np.array([
        [
            [1.0, 0.0, 0.0],
            [0.9, -0.3, 0.0],
            [0.8, -0.7, 0.1],
            [1.0, -1.0, 0.0]
        ],
        [
            [2.0, 0.0, 0.0],
            [2.1, -0.3, 0.1],
            [1.9, -0.7, 0.0],
            [2.0, -1.1, 0.0]
        ],
        [
            [2.5, 0.0, 0.0],
            [2.5, -0.4, 0.0],
            [2.5, -0.6, 0.0],
            [2.5, -1.0, 0.0]
        ]
    ]))
    bez_surf_target = ac.BezierSurface(np.zeros((4, 4, 3)))
    bez_surf_target.enforce_g0g1_multiface(
        adjacent_surf_u0=bez_surf_u0,
        adjacent_surf_u1=bez_surf_u1,
        adjacent_surf_v0=bez_surf_v0,
        adjacent_surf_v1=bez_surf_v1,
        other_edge_u0=ac.SurfaceEdge.u0,
        other_edge_u1=ac.SurfaceEdge.u0,
        other_edge_v0=ac.SurfaceEdge.u0,
        other_edge_v1=ac.SurfaceEdge.u0
    )

    surfs = [bez_surf_u0, bez_surf_u1, bez_surf_v0, bez_surf_v1, bez_surf_target]
    colors = ["blue", "red", "yellow", "purple", "green"]
    plot = pv.Plotter(off_screen=True if image_dir else False, window_size=[1024, 600])
    for idx, (surf, color) in enumerate(zip(surfs, colors)):
        surf.plot_surface(
            plot,
            50,
            50,
            color=color,
            show_edges=True if idx == 4 else False
        )
        surf.plot_control_points(
            plot,
            render_points_as_spheres=True,
            point_size=16,
            color="lime" if idx == 4 else "black"
        )
        surf.plot_control_point_mesh_lines(
            plot,
            color="gray"
        )
    plot.add_axes()
    plot.zoom_camera(3.0)
    if image_dir is not None:
        plot.screenshot(os.path.join(image_dir, "bezier_enforce_g0g1_multiface.png"), scale=1)
    else:
        plot.show()


def main(*args, **kwargs):
    # plot_bezier_surf_2x3(*args, **kwargs)
    # plot_bezier_surf_2x3_uv_labels(*args, **kwargs)
    # plot_bezier_surf_2x3_mesh_only(*args, **kwargs)
    # plot_bezier_surf_2x3_u_elevated(*args, **kwargs)
    # plot_bezier_surf_2x3_v_elevated(*args, **kwargs)
    # plot_fill_surface(*args, **kwargs)
    plot_enforce_g0g1_multiface(*args, **kwargs)


if __name__ == "__main__":
    # main()
    main(image_dir=r"C:\Users\mlaue\Documents\aerocaps\docs\source\images")



from aerocaps.units.unit import Unit


__all__ = [
    "Area"
]


area_conversions_from_m2 = {
    'ft2': (1 / 0.3048) ** 2,
    'mm2': 1000 ** 2,
    'cm2': 100 ** 2,
    'in2': (12 / 0.3048) ** 2,
}


class Area(Unit):
    def __init__(self, m2=None, mm2=None, cm2=None, ft2=None, in2=None):
        self._m2, self._mm2, self._cm2, self._ft2, self._in2 = None, None, None, None, None
        if m2 is not None:
            self.m2 = m2
        elif mm2 is not None:
            self.mm2 = mm2
        elif cm2 is not None:
            self.cm2 = cm2
        elif ft2 is not None:
            self.ft2 = ft2
        elif in2 is not None:
            self.in2 = in2
        super().__init__(primary_unit="m2")

    def set_all(self):
        """After setting the area in m2, set the area for all other units"""
        for k, v in area_conversions_from_m2.items():
            setattr(self, f'_{k}', self._m2 * v)

    @property
    def m2(self):
        return self._m2

    @m2.setter
    def m2(self, value):
        self._m2 = value
        self.set_all()

    @property
    def mm2(self):
        return self._mm2

    @mm2.setter
    def mm2(self, value):
        self._m2 = value / area_conversions_from_m2['mm2']
        self.set_all()

    @property
    def cm2(self):
        return self._cm2

    @cm2.setter
    def cm2(self, value):
        self._m2 = value / area_conversions_from_m2['cm2']
        self.set_all()

    @property
    def ft2(self):
        return self._ft2

    @ft2.setter
    def ft2(self, value):
        self._m2 = value / area_conversions_from_m2['ft2']
        self.set_all()

    @property
    def in2(self):
        return self._in2

    @in2.setter
    def in2(self, value):
        self._m2 = value / area_conversions_from_m2['in2']
        self.set_all()



import time
import typing

import matplotlib.pyplot as plt
import numpy as np
import numpy.random as rand
import rust_nurbs

import aerocaps as ac
from aerocaps.geom import nurbs_purepython


def case_1(N: int) -> (typing.List[float], str):
    """Bernstein polynomial evaluation at many random :math:`t`-values"""
    ts = rand.uniform(0.0, 1.0, N)
    start_py = time.perf_counter()
    for t in ts:
        nurbs_purepython.bernstein_poly(10, 4, t)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    for t in ts:
        rust_nurbs.bernstein_poly(10, 4, t)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 1"


def case_2(N: int) -> (typing.List[float], str):
    """Bézier curve evaluation"""
    ts = rand.uniform(0.0, 1.0, N)
    P = np.array([[0.0, 0.0, 0.0], [0.3, 0.5, 0.0], [0.1, -0.2, 0.3], [0.5, 0.1, 0.2], [0.6, 1.0, 2.0]])
    P_list = P.tolist()
    start_py = time.perf_counter()
    for t in ts:
        nurbs_purepython.bezier_curve_eval(P_list, t)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    for t in ts:
        rust_nurbs.bezier_curve_eval(P, t)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 2"


def case_3(N: int) -> (typing.List[float], str):
    """B-spline curve evaluation"""
    ts = rand.uniform(0.0, 1.0, N)
    P = np.array([[0.0, 0.0, 0.0], [0.3, 0.5, 0.0], [0.1, -0.2, 0.3], [0.5, 0.1, 0.2], [0.6, 1.0, 2.0]])
    P_list = P.tolist()
    knots = np.array([0.0, 0.0, 0.0, 0.0, 0.5, 1.0, 1.0, 1.0, 1.0])
    knot_list = knots.tolist()
    start_py = time.perf_counter()
    for t in ts:
        nurbs_purepython.bspline_curve_eval(P_list, knot_list, t)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    for t in ts:
        rust_nurbs.bspline_curve_eval(P, knots, t)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 3"


def case_4(N: int) -> (typing.List[float], str):
    """Bezier surface evaluation at a bunch of :math:`(u,v)`-pairs"""
    us = rand.uniform(0.0, 1.0, N)
    vs = rand.uniform(0.0, 1.0, N)
    P = np.array([
        [
            [0.0, 0.0, 0.0],
            [0.3, 0.5, 0.0],
            [0.1, -0.2, 0.3],
            [0.5, 0.1, 0.2],
            [0.6, 1.0, 2.0]
        ],
        [
            [0.0, 1.0, 0.5],
            [0.3, 1.5, 0.3],
            [0.1, 0.8, 0.3],
            [0.5, 1.1, 0.6],
            [0.6, 2.0, 3.0]
        ]
    ])
    P_list = P.tolist()
    start_py = time.perf_counter()
    for (u, v) in zip(us, vs):
        nurbs_purepython.bezier_surf_eval(P_list, u, v)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    for (u, v) in zip(us, vs):
        rust_nurbs.bezier_surf_eval(P, u, v)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 4"


def case_5(N: int) -> (typing.List[float], str):
    """High-resolution Bezier surface grid evaluation"""
    P = np.array([
        [
            [0.0, 0.0, 0.0],
            [0.3, 0.5, 0.0],
            [0.1, -0.2, 0.3],
            [0.5, 0.1, 0.2],
            [0.6, 1.0, 2.0]
        ],
        [
            [0.0, 1.0, 0.5],
            [0.3, 1.5, 0.3],
            [0.1, 0.8, 0.3],
            [0.5, 1.1, 0.6],
            [0.6, 2.0, 3.0]
        ]
    ])
    P_list = P.tolist()
    Nu, Nv = N, N
    start_py = time.perf_counter()
    nurbs_purepython.bezier_surf_eval_grid(P_list, Nu, Nv)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    rust_nurbs.bezier_surf_eval_grid(P, Nu, Nv)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 5"


def case_6(N: int) -> (typing.List[float], str):
    """High-resolution NURBS surface of revolution grid evaluation"""
    bez_P = np.array([
        [0.0, 1.0, 0.0],
        [0.2, 1.5, 0.0],
        [0.5, 0.8, 0.0],
        [0.8, 1.2, 0.0],
        [1.0, 1.1, 0.0]
    ])
    bez = ac.BezierCurve3D(bez_P)
    ax = ac.Line3D(p0=ac.Point3D.from_array(np.array([0.0, 0.0, 0.0])),
                   p1=ac.Point3D.from_array(np.array([1.0, 0.0, 0.0])))
    Nu, Nv = N, N
    surf = ac.NURBSSurface.from_bezier_revolve(bez, ax, start_angle=ac.Angle(deg=0.0), end_angle=ac.Angle(deg=360.0))
    P = surf.get_control_point_array()
    P_list = P.tolist()
    w = surf.weights
    w_list = w.tolist()
    ku = surf.knots_u
    ku_list = ku.tolist()
    kv = surf.knots_v
    kv_list = kv.tolist()

    start_py = time.perf_counter()
    nurbs_purepython.nurbs_surf_eval_grid(P_list, w_list, ku_list, kv_list, Nu, Nv)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    rust_nurbs.nurbs_surf_eval_grid(P, w, ku, kv, Nu, Nv)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 6"


def case_7(N: int) -> (typing.List[float], str):
    """High-resolution NURBS surface of revolution grid evaluation with higher degree in the :math:`v`-direction"""
    bez_P = np.array([
        [0.0, 1.0, 0.0],
        [0.2, 1.5, 0.0],
        [0.3, 1.2, 0.0],
        [0.4, 1.3, 0.0],
        [0.5, 0.8, 0.0],
        [0.6, 0.9, 0.0],
        [0.7, 0.8, 0.0],
        [0.8, 1.2, 0.0],
        [0.9, 1.1, 0.0],
        [1.0, 1.1, 0.0]
    ])
    bez = ac.BezierCurve3D(bez_P)
    ax = ac.Line3D(p0=ac.Point3D.from_array(np.array([0.0, 0.0, 0.0])),
                   p1=ac.Point3D.from_array(np.array([1.0, 0.0, 0.0])))
    Nu, Nv = N, N
    surf = ac.NURBSSurface.from_bezier_revolve(bez, ax, start_angle=ac.Angle(deg=0.0), end_angle=ac.Angle(deg=360.0))
    P = surf.get_control_point_array()
    P_list = P.tolist()
    w = surf.weights
    w_list = w.tolist()
    ku = surf.knots_u
    ku_list = ku.tolist()
    kv = surf.knots_v
    kv_list = kv.tolist()

    start_py = time.perf_counter()
    nurbs_purepython.nurbs_surf_eval_grid(P_list, w_list, ku_list, kv_list, Nu, Nv)
    end_py = time.perf_counter()

    start_rust = time.perf_counter()
    rust_nurbs.nurbs_surf_eval_grid(P, w, ku, kv, Nu, Nv)
    end_rust = time.perf_counter()

    return [end_py - start_py, end_rust - start_rust], "Case 7"


def main():

    sizes = [500000, 500000, 500000, 500000, 500, 500, 500]
    case_names = []
    time_arr = []
    for i, size in enumerate(sizes, start=1):
        times, name = globals()[f"case_{i}"](size)
        print(f"Completed {name}. Pure-Python: {times[0]:.3f} seconds. Rust-Python: {times[1]:.3f} seconds.")
        case_names.append(name)
        time_arr.append(times)

    relative_times = [[t[0] / t[1], 1.0] for t in time_arr]

    x = np.arange(len(case_names))
    width = 0.35

    fig, ax = plt.subplots()
    plt.grid(axis="y", ls=":")
    ax.bar(x - width / 2, [t[0] for t in relative_times], width, label="Pure Python")
    ax.bar(x + width / 2, [t[1] for t in relative_times], width, label="Rust-Python")

    ax.set_ylabel("Time Relative to Rust-Python", font=dict(size=16))
    for tick in ax.get_yticklabels():
        tick.set_fontsize(14)
    ax.set_xticks(x)
    ax.set_xticklabels(case_names, font=dict(size=14))
    ax.legend(prop=dict(size=16))

    fig.set_tight_layout(True)
    # fig.savefig("python-rust-speed-comparison.png", dpi=300, bbox_inches="tight")

    plt.show()


if __name__ == "__main__":
    main()
